import cocotb
from cocotb.clock import Clock
from cocotb.triggers import Timer, RisingEdge

import logging

from decimal import Decimal

from dataclasses import dataclass

import random

CLK_PERIOD_NS = 10

@dataclass()
class tcb:
    local_socket: int = 0
    remote_socket: int = 0
    send_ptr: int = 0
    recv_ptr: int = 0
    retransmit_ptr: int = 0
    user_timer: int = 0
    retransmit_timer: int = 0
    time_wait_timer: int = 0
    state: int = 0

    def pack(self):
        return self.local_socket

    def randomize(self):
        self.local_socket = random.randint(1, 65535)
        self.remote_socket = random.randint(1, 65535)

        return self

class TB:
    def __init__(self, dut):
        self.dut = dut

        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.INFO)

        cocotb.start_soon(Clock(dut.clk, CLK_PERIOD_NS, units="ns").start())

    async def cycle_reset(self):
        self.dut.rst.setimmediatevalue(0)
        await RisingEdge(self.dut.clk)  # type: ignore
        await RisingEdge(self.dut.clk)  # type: ignore
        self.dut.rst.value = 1
        await RisingEdge(self.dut.clk)  # type: ignore
        await RisingEdge(self.dut.clk)  # type: ignore
        self.dut.rst.value = 0
        await RisingEdge(self.dut.clk)  # type: ignore
        await RisingEdge(self.dut.clk)  # type: ignore


    async def tcb_mem_write(self, addr: int, data: tcb):
        self.dut.w_tcb_addr.value = addr
        self.dut.w_tcb.value = data.pack()
        self.dut.w_tcb_valid.value = 1
        await RisingEdge(self.dut.clk) # type: ignore
        self.dut.w_tcb_valid.value = 0


@cocotb.test()
async def sanity(dut):
    tb = TB(dut)
    await tb.cycle_reset()

    tcbs = [tcb().randomize() for _ in range(16)]

    for addr, _tcb in enumerate(tcbs):
        await tb.tcb_mem_write(addr, _tcb)


    tb.dut.r_tcb_ready.value = 1

    seen = 0
    addrs_seen = {i: 0 for i in range(16)}

    while True:
        await RisingEdge(tb.dut.clk) # type: ignore
        if tb.dut.r_tcb_valid.value:
            seen+=1
            addr = tb.dut.r_tcb_addr.value.integer
            _tcb = tb.dut.r_tcb.value.integer
            expected = tcbs[tb.dut.r_tcb_addr.value.integer].pack()
            addrs_seen[addr] += 1
            tb.log.info(f"Addr: {addr:x}")
            tb.log.info(f"Tcb: {_tcb:x} Expected: {expected:x}")
            assert _tcb == expected
        if seen == 64:
            break

    tcbs[4].randomize()
    await tb.tcb_mem_write(4, tcbs[4])

    seen = 0
    addrs_seen = {i: 0 for i in range(16)}

    while True:
        await RisingEdge(tb.dut.clk) # type: ignore
        if tb.dut.r_tcb_valid.value:
            seen+=1
            addr = tb.dut.r_tcb_addr.value.integer
            _tcb = tb.dut.r_tcb.value.integer
            expected = tcbs[tb.dut.r_tcb_addr.value.integer].pack()
            addrs_seen[addr] += 1
            tb.log.info(f"Addr: {addr:x}")
            tb.log.info(f"Tcb: {_tcb:x} Expected: {expected:x}")
            assert _tcb == expected
        if seen == 64:
            break

    tb.log.info(addrs_seen)